/***********************************************************************

  consoleFcns.c - console interface functions

  author    John Madsen
  revision  1.0
  date      2022-07-25


  history   1.0 initial release

***********************************************************************/


#include "consoleFcns.h"
#include "consoleCmds.h"
#include "storageFcns.h"
#include "systemFcns.h"

#include <stdio.h>
#include <string.h>
#include "esp_system.h"
#include "esp_log.h"
#include "esp_console.h"
#include "esp_vfs_dev.h"
#include "esp_vfs_fat.h"
#include "nvs.h"
#include "nvs_flash.h"
#include "linenoise/linenoise.h"
#include "argtable3/argtable3.h"



typedef unsigned int size_t;


static void init_filesystem(void);
static void init_nvs(void);



static const char			*Tag = "console";
static wl_handle_t 			wl_handle;
static TaskHandle_t 		xHandle = NULL;



/*******************************************************************

		Public Functions

 *******************************************************************/

// runs when USB plugged in (charger or terminal)
void consoleInitTask(void *pvParameter)
{
	consoleInit();
	storageLogWriteEvent(EL_USB_CONSOLE_CONNECT, SUCCESS);
	vTaskDelete(xHandle);
}



void consoleRunInit(void)
{
	xTaskCreate(&consoleInitTask, "console_init_task", 4096, NULL, 4, &xHandle);
}



void consoleInit(void)
{
    esp_console_repl_t *Repl = NULL;
    esp_console_repl_config_t repl_config = ESP_CONSOLE_REPL_CONFIG_DEFAULT();

    repl_config.prompt = PROMPT_STR ">";
    repl_config.max_cmdline_length = CONFIG_CONSOLE_MAX_COMMAND_LINE_LENGTH;


    init_nvs();
    init_filesystem();
    repl_config.history_save_path = HISTORY_PATH;
    repl_config.task_stack_size = 16384;
    ESP_LOGI(Tag, "Command history enabled");


    /* Register commands */
    esp_console_register_help_command();
//    register_system_common();
    register_storage_cmds();
//    register_sysdata_cmds();
//    register_fpsnsr_cmds();
//    register_cap_cmds();
    ESP_LOGI(Tag, "Registered commands");

    esp_console_dev_usb_serial_jtag_config_t hw_config = {};//ESP_CONSOLE_DEV_USB_SERIAL_JTag_CONFIG_DEFAULT();
    ESP_ERROR_CHECK(esp_console_new_repl_usb_serial_jtag(&hw_config, &repl_config, &Repl));
    ESP_LOGI(Tag, "esp_console_new_repl_usb_serial_jtag");


    ESP_ERROR_CHECK(esp_console_start_repl(Repl));
    ESP_LOGI(Tag, "esp_console_start_repl");
}



/*******************************************************************

		Private Functions

 *******************************************************************/

static void init_filesystem(void)
{
	esp_err_t ret;
    const esp_vfs_fat_mount_config_t mount_config =
    {
        .max_files = 4,
        .format_if_mount_failed = true
    };


    ESP_LOGI(Tag, "Mounting spiflash filesystem");
    ret = esp_vfs_fat_spiflash_mount(MOUNT_PATH, "storage", &mount_config, &wl_handle);

    if (ret == ESP_OK)
    	ESP_LOGI(Tag, "spiflash Filesystem mounted");
    else
    	ESP_LOGE(Tag, "Failed to mount spiflash filesystem (%s)", esp_err_to_name(ret));
}


static void init_nvs(void)
{
    esp_err_t err = nvs_flash_init();
    if (err == ESP_ERR_NVS_NO_FREE_PAGES || err == ESP_ERR_NVS_NEW_VERSION_FOUND) {
        ESP_ERROR_CHECK( nvs_flash_erase() );
        err = nvs_flash_init();
    }
    ESP_ERROR_CHECK(err);
}







