/***********************************************************************

  storageFcns.c - SD card functions

  author    John Madsen
  revision  1.0
  date      2022-05-23


  history   1.0 initial release

***********************************************************************/

#include "storageFcns.h"
#include "systemFcns.h"
#include "consoleFcns.h"
#include "ymodem.h"
#include "miniz.h"

#include "mytypes.h"
#include <string.h>
#include <sys/unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <fcntl.h>
#include "esp_vfs_fat.h"
#include "sdmmc_cmd.h"
#include "driver/sdmmc_host.h"
#include "ff.h"
#include "vfs_fat_internal.h"
#include "esp_camera.h"
#include "img_converters.h"
#include "esp_system.h"





typedef unsigned int size_t;

static const char 						*Tag = "storage";
static sdmmc_card_t 					*Card;
static const char 						MountPoint[] = MOUNT_POINT;
static const char 						*FileSystemLog = MOUNT_POINT"/systemlog.txt";
static esp_vfs_fat_sdmmc_mount_config_t	MountConfig;
static sdmmc_host_t 					Host = SDMMC_HOST_DEFAULT();
sdmmc_slot_config_t 					SlotConfig = SDMMC_SLOT_CONFIG_DEFAULT();
static monitorlog_dest_t				Destination = LD_CONSOLE;
RTC_NOINIT_ATTR int						LogIndex;
RTC_NOINIT_ATTR int						PicIndex;




/*******************************************************************

		Public Functions

 *******************************************************************/

void storageInit(void)
{
    esp_err_t ret;

    // set queue name
//    xQueueFrameI = frame_i;

    // Options for mounting the filesystem.
//    MountConfig =
//    {
	MountConfig.format_if_mount_failed = false;
	MountConfig.max_files = 5;
	MountConfig.allocation_unit_size = 16 * 1024;
//    };

    ESP_LOGI(Tag, "Initializing SD Card");

    // Use settings defined above to initialize SD Card and mount FAT filesystem.
    // Note: esp_vfs_fat_sdmmc/sdspi_mount is all-in-one convenience functions.
    // Please check its source code and implement error recovery when developing
    // production applications.

    ESP_LOGI(Tag, "Using SDMMC peripheral");
//    sdmmc_host_t Host = SDMMC_HOST_DEFAULT();

    // This initializes the slot without Card detect (CD) and write protect (WP) signals.
    // Modify SlotConfig.gpio_cd and SlotConfig.gpio_wp if your board has these signals.
//    sdmmc_slot_config_t SlotConfig = SDMMC_SLOT_CONFIG_DEFAULT();

    // interface width
    SlotConfig.width = SDMMC_WIDTH_1;

    // signals
    SlotConfig.clk = SDMMC_CLK;
    SlotConfig.cmd = SDMMC_CMD;
    SlotConfig.d0 = SDMMC_D0;
    SlotConfig.d1 = SDMMC_D1;
    SlotConfig.d2 = SDMMC_D2;
    SlotConfig.d3 = SDMMC_D3;

    // Enable internal pullups on enabled pins. The internal pullups
    // are insufficient however, please make sure 10k external pullups are
    // connected on the bus. This is for debug / example purpose only.
//    SlotConfig.flags |= SDMMC_SLOT_FLAG_INTERNAL_PULLUP;


    ESP_LOGI(Tag, "Mounting SD filesystem");
    ret = esp_vfs_fat_sdmmc_mount(MountPoint, &Host, &SlotConfig, &MountConfig, &Card);

    if (ret != ESP_OK) {
        if (ret == ESP_FAIL) {
            ESP_LOGE(Tag, "Failed to mount filesystem. "
                     "If you want the Card to be formatted, set the EXAMPLE_FORMAT_IF_MOUNT_FAILED menuconfig option.");
        } else {
            ESP_LOGE(Tag, "Failed to initialize the Card (%s). "
                     "Make sure SD Card lines have pull-up resistors in place.", esp_err_to_name(ret));
        }
        return;
    }
    ESP_LOGI(Tag, "SD Filesystem mounted");

    // Card has been initialized, print its properties
    sdmmc_card_print_info(stdout, Card);
}



void storageReadDir()
{
	DIR 			*dir;
	struct dirent	*entry;
	char 			type;
    char 			size[12];
    char 			path[FILE_PATH_MAX];
	struct stat 	stats;
    char 			tbuffer[80];
    struct 			tm *tm_info;
    int 			statok;
    uint64_t 		total = 0;
    int 			nfiles = 0;



	dir = opendir(MOUNT_POINT);
    if (!dir)
    {
    	ESP_LOGE(Tag, "Error opening directory %s", MOUNT_POINT);
        return;
    }

    printf("\nT     Size      Date/Time      Name\r\n");
    printf("-----------------------------------\r\n");
    while ((entry = readdir(dir)) != NULL)
    {
    	snprintf(path, sizeof(path), MOUNT_POINT"/%s", entry->d_name);

        statok = stat(path, &stats);

		if (statok == 0)
		{
			tm_info = localtime(&stats.st_mtime);
			strftime(tbuffer, 80, "%d/%m/%Y %R", tm_info);
		}
		else
			sprintf(tbuffer, "                ");

		if (entry->d_type == DT_REG)
		{
			type = 'f';
			nfiles++;
			if (statok)
				strcpy(size, "       ?");
			else
			{
				total += stats.st_size;
				if (stats.st_size < (1024*1024))
					sprintf(size,"%8d", (int)stats.st_size);
				else if ((stats.st_size/1024) < (1024*1024))
					sprintf(size,"%6dKB", (int)(stats.st_size / 1024));
				else
					sprintf(size,"%6dMB", (int)(stats.st_size / (1024 * 1024)));
			}
		}
		else
		{
			type = 'd';
			strcpy(size, "       -");
		}

		printf("%c  %s  %s  %s\r\n", type, size, tbuffer, entry->d_name);
    }

    if (total)
    {
        printf("-----------------------------------\r\n");
    	if (total < (1024*1024))
    		printf("   %7dB", (int)total);
    	else if ((total/1024) < (1024*1024))
    		printf("   %6dKB", (int)(total / 1024));
    	else
    		printf("   %6dMB", (int)(total / (1024 * 1024)));
    	printf(" in %d file(s)\r\n", nfiles);
    }
    printf("-----------------------------------\r\n\n");

    closedir(dir);
}



int storageSendFile(char *filename)
{
    FILE 			*fp = NULL;
    int				trans_res=-1;
    char 			path[FILE_PATH_MAX];
	struct stat 	stats;
    int 			statok;


	snprintf(path, sizeof(path), MOUNT_POINT"/%s", filename);

    statok = stat(path, &stats);
	if (statok)
	{
		ESP_LOGE(Tag, "File not found");
		return -1;
	}

//	ESP_LOGI(Tag, "Send file %s.", filename);
//	storageLogWriteEvent(EL_DATA_TRANFER_INIT, SUCCESS);

	fp = fopen(path, "rb");
	if (fp)
	{
		printf("\r\nSending file \"%s\", please start YModem receive on host ...\r\n", path);
		trans_res = Ymodem_Transmit(filename, stats.st_size, fp);
		fclose(fp);
		printf("\r\n");
//		if (trans_res == 0)
//		{
//			ESP_LOGI(Tag, "Transfer complete.");
//			storageLogWriteEvent(EL_DATA_TRANSFER_COMPL, SUCCESS);
//		}
//		else
//		{
//			ESP_LOGE(Tag, "Transfer complete, Error=%d", trans_res);
//			storageLogWriteEvent(EL_DATA_TRANSFER_COMPL, trans_res);
//		}
	}
//	else
//	{
//		ESP_LOGE(Tag, "Error opening file \"%s\" for sending.", path);
//		storageLogWriteEvent(EL_DATA_TRANSFER_COMPL, FAILURE);
//	}

	return 0;
}






int storprintf(const char *fmt, ...)
{
    static bool static_fatal_error = false;
    int result, iresult;
    va_list args;



    // always write to stdout
    va_start(args, fmt);
    result = vprintf(fmt, args);


    // Write to SD Card
    if (Destination == LD_FILE)
    {
		FILE *f = fopen(FileSystemLog, "a");
		if (f == NULL)
		{
//			printf("%s() ABORT. file handle fpFileSystemLog is NULL\n", __FUNCTION__);
			printf("%s() ABORT. file handle for %s is NULL\n", __FUNCTION__, FileSystemLog);
			result = -1;
//			return -1;
		}
		else if (static_fatal_error == false)
		{
			iresult = vfprintf(f, fmt, args);
			if (iresult < 0)
			{
				printf("%s() ABORT. failed vfprintf() -> disable future vfprintf(fpFileSystemLog) \n", __FUNCTION__);
				// MARK FATAL
				static_fatal_error = true;
				result = iresult;
//				return iresult;
			}
		}

		fclose(f);
    }

    va_end(args);
    return result;
}




/*******************************************************************

		Private Functions

 *******************************************************************/




