/***********************************************************************

 systemFcns.c - GPIO functions

  author    John Madsen
  revision  1.0
  date      2022-05-19


  history   1.0 initial release

***********************************************************************/


#include "systemFcns.h"
#include "io_expander.h"
#include "storageFcns.h"
#include "consoleFcns.h"

#include <stdio.h>
#include <sys/time.h>

#include "sdkconfig.h"
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "esp_system.h"
#include "esp_spi_flash.h"
#include "esp_log.h"
#include "esp_sleep.h"
#include "esp_console.h"
#include "driver/gpio.h"
#include "esp_adc_cal.h"





static time_t 				StartTime, ActivateTime;
static const char 			*Tag = "system";
static wake_cause_t			WakeCause;
static esp_reset_reason_t 	ResetReason;
static uint8_t				MacAddrBT[6];
static uint8_t				IOPortData;

SemaphoreHandle_t 				systask_mutex;
RTC_NOINIT_ATTR  int 			BootCount;
RTC_NOINIT_ATTR  stateflags_t	Flags;


/*******************************************************************

		Public Functions

 *******************************************************************/

void systemInit(void)
{
	struct timeval current_time;
	esp_chip_info_t chip_info;
	char event[40];
	esp_sleep_wakeup_cause_t	wake_cause;

	/* Print chip information */
	esp_chip_info(&chip_info);
	storprintf("This is %s chip with %d CPU core(s), WiFi%s%s, ",
			CONFIG_IDF_TARGET,
			chip_info.cores,
			(chip_info.features & CHIP_FEATURE_BT) ? "/BT" : "",
			(chip_info.features & CHIP_FEATURE_BLE) ? "/BLE" : "");

	storprintf("silicon revision %d, ", chip_info.revision);

	storprintf("%dMB %s flash\n", spi_flash_get_chip_size() / (1024 * 1024),
			(chip_info.features & CHIP_FEATURE_EMB_FLASH) ? "embedded" : "external");

	// read MAC address
	esp_read_mac(MacAddrBT, ESP_MAC_BT);
	storprintf("MAC Address:  %x:%x:%x:%x:%x:%x\n", MacAddrBT[0], MacAddrBT[1], MacAddrBT[2], MacAddrBT[3], MacAddrBT[4], MacAddrBT[5]);
//	storprintf("Minimum free heap size: %d bytes\n", esp_get_minimum_free_heap_size());

	// System time
	time(&StartTime);
//	storprintf("StartTime = %ld\n", StartTime);
	ActivateTime = StartTime;
	gettimeofday(&current_time, NULL);
	storprintf("%s",asctime (localtime(&current_time.tv_sec)));
	storprintf("Boot Count = %d\n", ++BootCount);


	// Get and print Reset Reason
	ResetReason = esp_reset_reason();

	switch(ResetReason)
	{
		case ESP_RST_POWERON:
			BootCount = 0;
			systemClearFlags();
			snprintf(event, sizeof(event), "Reset:  Power On");
			break;

		case ESP_RST_EXT:
			snprintf(event, sizeof(event), "Reset:  External Pin");
			break;

		case ESP_RST_SW:
			snprintf(event, sizeof(event), "Reset:  Software");
			break;

		case ESP_RST_PANIC:
			snprintf(event, sizeof(event), "Reset:  Panic");
			break;

		case ESP_RST_DEEPSLEEP:
			snprintf(event, sizeof(event), "Reset:  Deep Sleep");
			break;

		case ESP_RST_BROWNOUT:
			snprintf(event, sizeof(event), "Reset:  Brownout");
			break;

		case ESP_RST_UNKNOWN:		// after FW download
			BootCount = 0;
			systemClearFlags();
			snprintf(event, sizeof(event), "Reset:  Unknown Reset");
			break;

		default:
			snprintf(event, sizeof(event), "Other reset reason");
	}
//	storprintf("%s\n", event);
//	storageLogWriteEvent(EL_RESET_REASON, ResetReason);


	// Get and print Wake Up cause
	wake_cause = esp_sleep_get_wakeup_cause();

	switch(wake_cause)
	{
		case ESP_SLEEP_WAKEUP_EXT0:
			if (IOPortData & IO_USB_CONN)
			{
				WakeCause = WC_USB_CONN;
				snprintf(event, sizeof(event), "Wake up:  USB");
			}
			else if (IOPortData & IO_FP_INT)
			{
				WakeCause = WC_FP_SENSOR;
				snprintf(event, sizeof(event), "Wake up:  FP Sensor");
			}
			else if (IOPortData & IO_LID_CLOSED)
			{
				WakeCause = WC_LID_CLOSE;
				snprintf(event, sizeof(event), "Wake up:  Lid Close");
			}
			else
			{
				WakeCause = WC_PORT_UNKNOWN;	// can be lid opening - bouncing switch, USB removal, falling edge of FP sensor interrupt
				snprintf(event, sizeof(event), "Wake up:  IO Exp Port Unknown");
			}
			break;

		case ESP_SLEEP_WAKEUP_TIMER:
			WakeCause = WC_TIMER;
			snprintf(event, sizeof(event), "Wake up:  Timer");
			break;

		case ESP_SLEEP_WAKEUP_UNDEFINED:
			WakeCause = WC_UNDEFINED;
			snprintf(event, sizeof(event), "Wake up:  Reset");
			break;

		default:
			WakeCause = WC_OTHER;
			snprintf(event, sizeof(event), "Other wake up cause");
	}

//	storprintf("%s\n", event);
//	storageLogWriteEvent(EL_WAKE_CAUSE, WakeCause);

	// for Phase 3 - look at wakup cause and only start what needs to be started.
	systemWake();
	// probably should keep looking for USB plug in, even if not cause of wake
}


void systemGetIOPort(void)
{
	IOPortData = ioExpanderReadInPort() & PORT_DIRECTIONS;
}





void systemWake(void)
{
//	wake_source_t	wake_source = WS_NULL;



    // create mutex
    systask_mutex = xSemaphoreCreateMutex();


		consoleInit();
}






/*******************************************************************

		Access Functions

 *******************************************************************/

wake_cause_t systemGetWakeCause(void)
{
	return WakeCause;
}

esp_reset_reason_t systemGetResetReason(void)
{
	return ResetReason;
}

int systemGetBootCount(void)
{
	return BootCount;
}

uint8_t* systemGetMACAddress(void)
{
	return MacAddrBT;
}

void systemGetFlags(void)
{
	printf ("Flags.collarLocked = %d\n", Flags.collarLocked);
	printf ("Flags.collarLockArm = %d\n", Flags.collarLockArm);
	printf ("Flags.lidReleased = %d\n", Flags.lidReleased);
	printf ("Flags.takePicture = %d\n", Flags.takePicture);
	printf ("Flags.userRegistered = %d\n", Flags.userRegistered);
}

void systemClearFlags(void)
{
	Flags.collarLocked = 0;
	Flags.collarLockArm = 0;
	Flags.lidReleased = 0;
	Flags.takePicture = 0;
	Flags.userRegistered = 0;
}



/*******************************************************************

		Private Functions

 *******************************************************************/




